/*
 * Copyright (c) 2012, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "lex_analyzer.h"
#include "CzFStream.h"
#include "CzarBasic.h"
#include "CzarBasic_error.h"

/* 演算子検索 */
static TOKEN_TYPE search_operator1(int ch);
static TOKEN_TYPE search_operator2(int ch1, int ch2);
/* 予約語検索 */
static TOKEN_TYPE search_keyword(char *token);


/* 記号チェック */
static int isMark(int ch);

static TOKEN st_token;
static int st_unget_token = 0;

enum token_status {
	STATUS_NONE,		/* 無し */
	STATUS_STRING,		/* 文字列 */
	STATUS_WORD,		/* 単語 */
};

/* 記号一覧 */
static const char *st_mark = "!\"#$%&'()*+,-./:;<=>?[\\]^_`{|}~";

/* キーワード */
static struct keyword {
	TOKEN_TYPE token_type;
	char *keyword;
} st_keyword[] = {
	{ TOKEN_TYPE_PRINT, "PRINT" },	/* PRINT */
	{ TOKEN_TYPE_IF, "IF" },	/* IF */
	{ TOKEN_TYPE_GOTO, "GOTO" },	/* GOTO */
	{ TOKEN_TYPE_LET, "LET" },		/* LET */
	{ TOKEN_TYPE_INPUT, "INPUT" },	/* INPUT */
	{ TOKEN_TYPE_GOSUB, "GOSUB" },	/* GOSUB */
	{ TOKEN_TYPE_RETURN, "RETURN" },	/* RETURN */
	{ TOKEN_TYPE_CLEAR, "CLEAR" },	/* CLEAR */
	{ TOKEN_TYPE_END, "END" },		/* END */
	{ TOKEN_TYPE_THEN, "THEN" },	/* THEN */
	{ TOKEN_TYPE_REM, "REM" },	/* REM */
	{ TOKEN_TYPE_REM, "'" },	/* REM */
	{ TOKEN_TYPE_NONE, NULL },
};

/* １文字の演算子および1文字記号のステートメント */
static struct operator_1 {
	TOKEN_TYPE token_type;
	char *keyword;
} st_operator_1[] = {
	{ TOKEN_TYPE_LPARENTHESES, "(" },
	{ TOKEN_TYPE_RPARENTHESES, ")" },
	{ TOKEN_TYPE_PLUS, "+" },
	{ TOKEN_TYPE_MINUS, "-" },
	{ TOKEN_TYPE_MUL, "*" },
	{ TOKEN_TYPE_DIV, "/" },
	{ TOKEN_TYPE_EQ, "=" },
	{ TOKEN_TYPE_LT, "<" },
	{ TOKEN_TYPE_GT, ">" },
	{ TOKEN_TYPE_REM, "'" },
	{ TOKEN_TYPE_COLON, ":" },
	{ TOKEN_TYPE_COMMA, "," },
	{ TOKEN_TYPE_SEMICOLON, ";" },
	{ TOKEN_TYPE_NONE, NULL },
};

/* ２文字の演算子 */
static struct operator_2 {
	TOKEN_TYPE token_type;
	char *keyword;
} st_operator_2[] = {
	{ TOKEN_TYPE_NE, "<>" },
	{ TOKEN_TYPE_LE, "<=" },
	{ TOKEN_TYPE_GE, ">=" },
	{ TOKEN_TYPE_NONE, NULL},
};

/* Tokenを取得する */
TOKEN *lex_get_token(LEX_DATA *stream)
{
	int ch;
	int pt;
	int ch2;
	enum token_status status;
	TOKEN_TYPE type;

	if(st_unget_token) {
		st_unget_token = 0;
		return &st_token;
	}

	pt = 0;
	status = STATUS_NONE;
	while((ch = stream->stream->getch(stream->stream)) != EOF) {
		if(pt >= STRING_SIZE - 1) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			st_token.token_type = TOKEN_TYPE_ERROR;
			return &st_token;
		}
		switch(status) {
		case STATUS_NONE:	/* 無し */
			if(ch == '\r') { /* CRなら */
				ch2 = stream->stream->getch(stream->stream);
				if(ch2 != '\n') { /* LF以外なら */
					/* 戻す */
					stream->stream->ungetc(stream->stream, ch2);
				}
				st_token.token_type = TOKEN_TYPE_NL;
				st_token.token[0] = 0;
				return &st_token;
			} else if(ch == '\n') { /* LFなら */
				st_token.token_type = TOKEN_TYPE_NL;
				st_token.token[0] = 0;
				return &st_token;
			} else if(ch == '"') { /* 文字列開始なら */
				status = STATUS_STRING;
				st_token.token_type = TOKEN_TYPE_STRING;
			} else if(ch != ' ' && !isMark(ch) && ch != '\t') { /* 記号以外なら */
				status = STATUS_WORD;
				st_token.token_type = TOKEN_TYPE_WORD;
				st_token.token[pt] = (char)ch;
				pt++;
			} else {
				if(ch > ' ') {
					TOKEN_TYPE ope;
					st_token.token_type = TOKEN_TYPE_MARK;
					st_token.token[pt] = (char)ch;
					pt++;
					ch2 = stream->stream->getch(stream->stream);
					if((ope = search_operator2(ch, ch2)) != TOKEN_TYPE_NONE) {
						st_token.token_type = ope;
						st_token.token[pt] = (char)ch2;
						pt++;
					} else {
						stream->stream->ungetc(stream->stream, ch2);
						if((ope = search_operator1(ch)) != TOKEN_TYPE_NONE) {
							st_token.token_type = ope;
						}
					}
					st_token.token[pt] = 0;
					return &st_token;
				}
			}
			break;
		case STATUS_WORD:		/* 単語 */
			if(isdigit(ch) || isalpha(ch) || ch == '.') { /* 数値またはアルファベットなら */
				st_token.token[pt] = (char)ch;
				pt++;
			} else {
				/* 戻す */
				stream->stream->ungetc(stream->stream, ch);
				st_token.token_type = TOKEN_TYPE_WORD;
				st_token.token[pt] = 0;
				type = search_keyword(st_token.token);
				if(type != TOKEN_TYPE_NONE) {
					st_token.token_type = type;
				}
				return &st_token;
			}
			break;
		case STATUS_STRING:		/* 文字列 */
			if(ch != '"') {
				st_token.token[pt] = (char)ch;
				pt++;
			} else {
				st_token.token[pt] = 0;
				return &st_token;
			}
			break;
		}
	}

	if(pt == 0) {
		st_token.token_type = TOKEN_TYPE_EOF;
		st_token.token[0] = 0;
		return &st_token;
	}
	st_token.token[pt] = 0;

	return &st_token;
}

/* tokenを戻す */
void lex_unget_token(TOKEN *token)
{
	st_unget_token = 1;
	if(token != &st_token) {
		memcpy(&st_token, token, sizeof(TOKEN));
	}
}

/* １文字演算子を検索 */
static TOKEN_TYPE search_operator1(int ch)
{
	int i = 0;

	while(st_operator_1[i].token_type != TOKEN_TYPE_NONE) {
		if(st_operator_1[i].keyword[0] == (char)ch) {
			return st_operator_1[i].token_type;
		}
		i++;
	}
	return TOKEN_TYPE_NONE;
}

/* ２文字演算子を検索 */
static TOKEN_TYPE search_operator2(int ch1, int ch2)
{
	int i = 0;

	while(st_operator_2[i].token_type != TOKEN_TYPE_NONE) {
		if(st_operator_2[i].keyword[0] == (char)ch1 && 
				st_operator_2[i].keyword[1] == (char)ch2) {
			return st_operator_2[i].token_type;
		}
		i++;
	}
	return TOKEN_TYPE_NONE;
}

#ifdef LEX_TEST
int lex_test(int argc, char *argv[])
{
	TOKEN *token;
	CZFSTREAM stream;
	LEX_DATA lex_data;
	FILE *fp;

	fprintf(stderr, "LEX_TEST\n");
	fp = fopen(argv[1], "r");
	CzFStream_init(&stream, fp);
	lex_data_init(&lex_data, (CZSTREAM*)&stream);
	while((token = lex_get_token(&lex_data)) != NULL) {
		if(token->token_type == TOKEN_TYPE_EOF) {
			break;
		}
		printf("-----\n");
		printf("token_type:");
		switch(token->token_type) {
		case TOKEN_TYPE_NL:		/* 改行 */
			printf("改行\n");
			break;
		case TOKEN_TYPE_WORD:		/* 単語 */
			printf("単語\n");
			break;
		case TOKEN_TYPE_SPACE:	/* スペース */
			printf("スペース\n");
			break;
		case TOKEN_TYPE_MARK:	/* 記号 */
			printf("記号\n");
			break;
		case TOKEN_TYPE_STRING:	/* 文字列 */
			printf("文字列\n");
			break;
		case TOKEN_TYPE_LPARENTHESES:	/* ( */
			printf("(\n");
			break;
		case TOKEN_TYPE_RPARENTHESES:	/* ) */
			printf(")\n");
			break;
		case TOKEN_TYPE_PLUS:		/* + */
			printf("+\n");
			break;
		case TOKEN_TYPE_MINUS:		/* - */
			printf("-\n");
			break;
		case TOKEN_TYPE_MUL:		/* * */
			printf("*\n");
			break;
		case TOKEN_TYPE_DIV:		/* / */
			printf("/\n");
			break;
		case TOKEN_TYPE_EQ:		/* = */
			printf("=\n");
			break;
		case TOKEN_TYPE_GT:		/* > */
			printf(">\n");
			break;
		case TOKEN_TYPE_GE:		/* >= */
			printf(">=\n");
			break;
		case TOKEN_TYPE_LT:		/* < */
			printf("<\n");
			break;
		case TOKEN_TYPE_LE:		/* <= */
			printf("<=\n");
			break;
		case TOKEN_TYPE_NE:		/* <> */
			printf("<>\n");
			break;
		case TOKEN_TYPE_PRINT:	/* PRINT */
			printf("PRINT\n");
			break;
		case TOKEN_TYPE_IF:		/* IF */
			printf("PRINT\n");
			break;
		case TOKEN_TYPE_GOTO:	/* GOTO */
			printf("PRINT\n");
			break;
		case TOKEN_TYPE_LET:		/* LET */
			printf("LET\n");
			break;
		case TOKEN_TYPE_INPUT:	/* INPUT */
			printf("INPUT\n");
			break;
		case TOKEN_TYPE_GOSUB:	/* GOSUB */
			printf("GOSUB\n");
			break;
		case TOKEN_TYPE_RETURN:	/* RETURN */
			printf("RETURN\n");
			break;
		case TOKEN_TYPE_CLEAR:	/* CLEAR */
			printf("CLEAR\n");
			break;
		case TOKEN_TYPE_END:		/* END */
			printf("END\n");
			break;
		case TOKEN_TYPE_THEN:	/* THEN */
			printf("THEN\n");
			break;
		case TOKEN_TYPE_REM:	/* REM */
			printf("REM\n");
			break;
		default:
			printf("ERROR\n");
			break;
		}
		printf("token:'%s'\n", token->token);
	}
	stream.super.close((CZSTREAM*)&stream);
	lex_data_delete(&lex_data);
	fprintf(stderr, "LEX_TEST END\n");
	return 0;
}
#endif

/* 記号チェック */
static int isMark(int ch)
{
	UINT64 i;

	for(i = 0; i < strlen(st_mark); i++) {
		if(st_mark[i] == (char)ch) {
			return 1;
		}
	}
	return 0;
}

/* LEX_DATAの初期化 */
void lex_data_init(LEX_DATA *lex_data, CZSTREAM *stream)
{
	lex_data->stream = stream;
	lex_data->err_code = ERR_NOERROR;
	lex_data->err_line_num = -1;
	CzMemory_init(&lex_data->label_address);
}

/* LEX_DATAのメモリの開放 */
void lex_data_delete(LEX_DATA *lex_data)
{
	CzMemory_init(&lex_data->label_address);
}

/* 予約語検索 */
static TOKEN_TYPE search_keyword(char *token)
{
	int i;

	i = 0;
	while(st_keyword[i].token_type != TOKEN_TYPE_NONE) {
		if(Strcmp(st_keyword[i].keyword, token) == 0) {
			return st_keyword[i].token_type;
		}
		i++;
	}
	return TOKEN_TYPE_NONE;
}

/* 文字列比較 */
int Strcmp(const char *str1, const char *str2)
{
	UINT64 i;

	if(strlen(str1) != strlen(str2)) {
		return -1;
	}

	for(i = 0; i < strlen(str1); i++) {
		int ch1 = toupper((int)str1[i]);
		int ch2 = toupper((int)str2[i]);
		if(ch1 != ch2) {
			return -1;
		}
	}
	return 0;
}

