/*
 * Copyright (c) 2012, Masanori Miyamoto.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1.Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2.Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE FREEBSD PROJECT ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE FREEBSD PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 */ 
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "parser.h"
#include "lex_analyzer.h"
#include "CzarBasic_vm.h"
#include "CzarBasic_error.h"

/* tokenを符号無し整数に変換 */
static PARSE_RESULT conv_unsigned_int(TOKEN *token, int *retval);

/* 文字列プールポインタ */
static int st_string_pt;

/* 行解析 */
static PARSE_RESULT parse_line(VM_DATA *vm_data, LEX_DATA *stream);
/* 命令解析 */
static PARSE_RESULT parse_statement(VM_DATA *vm_data, LEX_DATA *stream);
/* 命令リスト解析 */
static PARSE_RESULT parse_statement_list(VM_DATA *vm_data, LEX_DATA *stream);
/* 式解析 */
static PARSE_RESULT parse_expression(VM_DATA *vm_data, LEX_DATA *stream);

static PARSE_RESULT parse_print(VM_DATA *vm_data, LEX_DATA *stream);
static PARSE_RESULT parse_if(VM_DATA *vm_data, LEX_DATA *stream);
static PARSE_RESULT parse_goto(VM_DATA *vm_data, LEX_DATA *stream);
static PARSE_RESULT parse_let(VM_DATA *vm_data, LEX_DATA *stream);
static PARSE_RESULT parse_input(VM_DATA *vm_data, LEX_DATA *stream);
static PARSE_RESULT parse_gosub(VM_DATA *vm_data, LEX_DATA *stream);
static PARSE_RESULT parse_return(VM_DATA *vm_data, LEX_DATA *stream);
static PARSE_RESULT parse_clear(VM_DATA *vm_data, LEX_DATA *stream);
static PARSE_RESULT parse_end(VM_DATA *vm_data, LEX_DATA *stream);
static PARSE_RESULT parse_rem(VM_DATA *vm_data, LEX_DATA *stream);
/* 項1解析 */
static PARSE_RESULT parse_item1(VM_DATA *vm_data, LEX_DATA *stream);
/* 項2解析 */
static PARSE_RESULT parse_item2(VM_DATA *vm_data, LEX_DATA *stream);
/* 基本式解析 */
static PARSE_RESULT parse_basis_expression(VM_DATA *vm_data, LEX_DATA *stream);

/* 命令の実行 */
static PARSE_RESULT parse_statement_call(VM_DATA *vm_data, TOKEN *token, LEX_DATA *stream);
/* 記号からオペコードへの変換 */
static enum opecode search_mark_to_opecode(TOKEN_TYPE mark);

/* 変数の解析　*/
static PARSE_RESULT parse_var(TOKEN *token, int *var_index);

/* ラベルに割り当てられているアドレスを変更 */
static void change_label_address(VM_DATA *vm_data, UINT32 label_index, UINT32 address);
/* ラベルの追加 */
static UINT32 add_label(VM_DATA *vm_data, char *label);

/* ラベル解析 */
static PARSE_RESULT parse_label(VM_DATA *vm_data, LEX_DATA *stream, int *index);

#define NL_STRING_PT	0
#define TAB_STRING_PT	1

typedef PARSE_RESULT (*STATEMENTFUNC)(VM_DATA *vm_data, LEX_DATA *);
static STATEMENTFUNC st_statement_func[] = {
	parse_print,
	parse_if,
	parse_goto,
	parse_let,
	parse_input,
	parse_gosub,
	parse_return,
	parse_clear,
	parse_end,
	NULL,		/* THEN */
	parse_rem,
};

static struct __mark_item {
	TOKEN_TYPE mark;	/* トークンタイプ */
	enum opecode opecode;	/* オペコード */
} st_mark_list[] = {
	{ TOKEN_TYPE_PLUS, OP_ADD },		/* + */
	{ TOKEN_TYPE_MINUS, OP_SUB },		/* - */
	{ TOKEN_TYPE_MUL, OP_MUL },		/* * */
	{ TOKEN_TYPE_DIV, OP_DIV },		/* / */
	{ TOKEN_TYPE_EQ, OP_CMP_EQ },		/* = */
	{ TOKEN_TYPE_GT, OP_CMP_GT },		/* > */
	{ TOKEN_TYPE_GE, OP_CMP_GE },		/* >= */
	{ TOKEN_TYPE_LT, OP_CMP_LT },		/* < */
	{ TOKEN_TYPE_LE, OP_CMP_LE },		/* <= */
	{ TOKEN_TYPE_NE, OP_CMP_NE },		/* <> */
	{ TOKEN_TYPE_NONE, OP_NOP },
};

/* 記号からオペコードへの変換 */
static enum opecode search_mark_to_opecode(TOKEN_TYPE mark)
{
	int i;
	i = 0;
	while(st_mark_list[i].mark != TOKEN_TYPE_NONE) {
		if(st_mark_list[i].mark == mark) {
			break;
		}
		i++;
	}
	return st_mark_list[i].opecode;
}
	

/* プログラムメモリへコード追加 */
static BOOL add_progmem(VM_DATA *vm_data, LEX_DATA *lex_data, int data)
{
	vm_data->mem[vm_data->pc] = data;
	vm_data->pc++;
	vm_data->mem_last = vm_data->pc;
	if(vm_data->pc >= VMMEM_MAX) {
		lex_data->err_code = ERR_PROGMEM_OVERFLOW;
		return FALSE;
	}
	return TRUE;
}

/* 文字列テーブルへの文字列追加 */
static BOOL add_string(VM_DATA *vm_data, LEX_DATA *lex_data, char *str)
{
	CZSTRING src, dest;
	size_t i;
	CZCHAR utf;

	CzString_init(&src);
	CzString_init(&dest);
	CzString_add_datas(&src, str, strlen(str) + 1);
#if defined(HAVE_ICONV) || defined(WIN32)
	if(!CzEncoding_convert(&vm_data->encMultbytetoUTF, &dest, &src)) {
		lex_data->err_code = ERR_OUTOFMEMORY;
		CzString_delete(&src);
		CzString_delete(&dest);
		return FALSE;
	}
#else
	for(i = 0; i < src.size; i++) {
		utf = src.data[i] & 0xff;
		if(!CzString_add_char(&dest, utf)) {
			lex_data->err_code = ERR_OUTOFMEMORY;
			CzString_delete(&src);
			CzString_delete(&dest);
			return FALSE;
		}
	}
#endif

	if(!CzMemory_add_datas(&vm_data->string_pool, (char*)&dest, sizeof(dest))) {
		lex_data->err_code = ERR_OUTOFMEMORY;
		CzString_delete(&src);
		CzString_delete(&dest);
		return FALSE;
	}
	CzString_delete(&src);
	st_string_pt++;

	return TRUE;
}

/* 行番号テーブルへ追加 */
static PARSE_RESULT add_line_num_table(VM_DATA *vm_data, LEX_DATA *stream, int address)
{
	if(vm_data->line_num_table_last >= LINE_NUM_MAX) {
		stream->err_code = ERR_LINENUMMEM_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}
	vm_data->line_num_table[vm_data->line_num_table_last] = address;
	vm_data->line_num_table_last++;

	return PARSE_RESULT_TRUE;
}

PARSE_RESULT parser(VM_DATA *vm_data, LEX_DATA *stream)
{
	size_t i;
	UINT32 *label_item, addr;

	st_string_pt = 0;
	/* 改行とタブを文字列テーブルに追加しておく */
	add_string(vm_data, stream, "\n");
	add_string(vm_data, stream, "\t");
	stream->err_line_num = 1;
	while(!stream->stream->eof(stream->stream)) {
		/* 行番号追加 */
		if(add_line_num_table(vm_data, stream, vm_data->pc) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
		/* 行解析 */
		if(parse_line(vm_data, stream) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
		stream->err_line_num++;
	}
	if(!add_progmem(vm_data, stream, OP_END)) {
		return PARSE_RESULT_FALSE;
	}
	/* ラベルの実際のアドレスを設定 */
	label_item = (UINT32*)stream->label_address.data;
	for(i = 0; i < stream->label_address.size / sizeof(UINT32); i++) {
		/* ラベルのアドレスを取得 */
		addr = get_label_address(vm_data, vm_data->mem[label_item[i]]);
		if(addr == (UINT32)-1) {
			stream->err_code = ERR_UNDEFINEDLABEL;
			stream->err_line_num = search_editor_line_num(vm_data, label_item[i]);
			return PARSE_RESULT_FALSE;
		}
		/* 実際のアドレスを設定 */
		vm_data->mem[label_item[i]] = addr;
	}
	return PARSE_RESULT_TRUE;
}

/* 行解析 */
static PARSE_RESULT parse_line(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN *token;
	int label_index;

	/* トークン取得 */
	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		stream->err_code = ERR_BUFFER_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}
	if(token->token_type == TOKEN_TYPE_EOF || token->token_type == TOKEN_TYPE_NL) { /* 改行、ファイル終了なら */
		return PARSE_RESULT_TRUE;
	}

	if(token->token[0] == '#') {	 /* コメント */
		/* 改行までスキップ */
		do {
			token = lex_get_token(stream);
			if(token->token_type == TOKEN_TYPE_ERROR) {
				stream->err_code = ERR_BUFFER_OVERFLOW;
				return PARSE_RESULT_FALSE;
			}
		} while(token->token_type != TOKEN_TYPE_NL &&
				token->token_type != TOKEN_TYPE_EOF);
	} else {
		int line_num;
		/* 行番号 */
		if(conv_unsigned_int(token, &line_num) == PARSE_RESULT_TRUE) {
			label_index = add_label(vm_data, token->token);
			/* 行番号をラベルテーブルに追加 */
			if(label_index == (UINT32)-1) {
				stream->err_code = ERR_OUTOFMEMORY;
				return PARSE_RESULT_FALSE;
			}
			/* ラベルにアドレスを割り当てる */
			change_label_address(vm_data, label_index, vm_data->pc);
		} else {
			line_num = -1;
			/* トークンを戻す */
			lex_unget_token(token);
		}

		/* ラベル */
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		if(token->token_type == TOKEN_TYPE_MUL) { /* '*'なら */
			if(parse_label(vm_data, stream, &label_index) == PARSE_RESULT_FALSE) {
				return PARSE_RESULT_FALSE;
			}
			/* ラベルにアドレスを設定 */
			change_label_address(vm_data, label_index, vm_data->pc);
		} else {
			/* トークンを戻す */
			lex_unget_token(token);
		}
		/* コロン(:) */
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		if(token->token_type != TOKEN_TYPE_COLON) { /* ':'以外なら */
			/* トークンを戻す */
			lex_unget_token(token);
		}

		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		if(token->token_type == TOKEN_TYPE_NL ||  token->token_type == TOKEN_TYPE_EOF) {
			return PARSE_RESULT_TRUE;
		}
		/* トークンを戻す */
		lex_unget_token(token);
		/* 命令リスト解析 */
		if(parse_statement_list(vm_data, stream) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
		/* CR */
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		if(token->token_type != TOKEN_TYPE_NL && token->token_type != TOKEN_TYPE_EOF) {
			stream->err_code = ERR_SYNTAX;
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* tokenを符号無し整数に変換 */
static PARSE_RESULT conv_unsigned_int(TOKEN *token, int *retval)
{
	int pt = 0;

	*retval = 0;
	if(token->token[0] == '0' && token->token[1] != 0) {
		return PARSE_RESULT_FALSE;
	}
	while(token->token[pt] != 0) {
		if(!isdigit((int)token->token[pt])) {
			return PARSE_RESULT_FALSE;
		}
		*retval = *retval * 10 + token->token[pt] - '0';
		if(*retval & ~0x7fff) {
			return PARSE_RESULT_OVERFLOW;
		}
		pt++;
	}
	return PARSE_RESULT_TRUE;
}


/* 命令解析 */
static PARSE_RESULT parse_statement(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN *token;
	PARSE_RESULT res;

	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		stream->err_code = ERR_BUFFER_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}
	/* 予約語解析して実行 */
	res = parse_statement_call(vm_data, token, stream);
	if(res == PARSE_RESULT_NOSTATEMENT) {	
		/* 予約語以外なら */
		lex_unget_token(token);
		/* 変数へ数値代入として解析する */
		return parse_let(vm_data, stream);
	}
	return res;
}

/* 命令ごとの解析ルーチンを呼ぶ */
static PARSE_RESULT parse_statement_call(VM_DATA *vm_data, TOKEN *token, LEX_DATA *stream)
{
	int func_no;

	func_no = token->token_type - TOKEN_TYPE_PRINT;

	if(func_no < 0) {
		return PARSE_RESULT_NOSTATEMENT;
	}

	if(st_statement_func[func_no]) {
		return st_statement_func[func_no](vm_data, stream);
	} else {
		stream->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_NOSTATEMENT;
}

/* PRINT解析 */
static PARSE_RESULT parse_print(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN *token;
	int add_nl = 0;

	while(1) {
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		if(token->token_type == TOKEN_TYPE_STRING) { /* 文字列なら */
			if(!(add_progmem(vm_data, stream, OP_PRN_STRING) &&
					add_progmem(vm_data, stream, st_string_pt) &&
					add_string(vm_data, stream, token->token))) {
				return PARSE_RESULT_FALSE;
			}
			add_nl = 1;
		} else if(token->token_type == TOKEN_TYPE_COMMA) { /* カンマ(,)なら */
			if(!(add_progmem(vm_data, stream, OP_PRN_STRING) &&
					add_progmem(vm_data, stream, TAB_STRING_PT))) {
				return PARSE_RESULT_FALSE;
			}
			add_nl = 1;
		} else if(token->token_type == TOKEN_TYPE_SEMICOLON) {
			add_nl = 0;
		} else if(token->token_type != TOKEN_TYPE_EOF && token->token_type != TOKEN_TYPE_NL && token->token_type != TOKEN_TYPE_COLON) {
			lex_unget_token(token);
			if(parse_expression(vm_data, stream) == PARSE_RESULT_FALSE) {
				return PARSE_RESULT_FALSE;
			}
			if(!add_progmem(vm_data, stream, OP_PRN)) {
				return PARSE_RESULT_FALSE;
			}
			add_nl = 1;
		}  else {
			lex_unget_token(token);
			break;
		}
	}

	if(add_nl) {
		if(!(add_progmem(vm_data, stream, OP_PRN_STRING) &&
				add_progmem(vm_data, stream, NL_STRING_PT))) {
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* IF解析 */
static PARSE_RESULT parse_if(VM_DATA *vm_data, LEX_DATA *stream)
{
	int jmp_addr;
	TOKEN *token;

	/* 式解析 */
	if(parse_expression(vm_data, stream) == PARSE_RESULT_FALSE) {
		return PARSE_RESULT_FALSE;
	}
	/* 偽ならJMPのオペコードを追加 */
	if(!add_progmem(vm_data, stream, OP_NCOND_JMP)) {
		return PARSE_RESULT_FALSE;
	}
	jmp_addr = vm_data->pc;
	/* 飛び先のアドレスを仮設定 */
	if(!add_progmem(vm_data, stream, 0)) {
		return PARSE_RESULT_FALSE;
	}
	
	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		stream->err_code = ERR_BUFFER_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}
	if(token->token_type != TOKEN_TYPE_THEN) { /* THENでなければ */
		stream->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	/* 次の命令を解析 */
	if(parse_statement_list(vm_data, stream) == PARSE_RESULT_FALSE) {
		return PARSE_RESULT_FALSE;
	}
	/* 偽のときの飛び先を再設定 */
	vm_data->mem[jmp_addr] = vm_data->pc;

	return PARSE_RESULT_TRUE;
}

/* GOTO解析 */
static PARSE_RESULT parse_goto(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN *token;
	int label_index;

	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		stream->err_code = ERR_BUFFER_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}
	if(token->token_type == TOKEN_TYPE_MUL) { /* ラベルなら */
		if(parse_label(vm_data, stream, &label_index) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
		if(!add_progmem(vm_data, stream, OP_JMP)) {
			return PARSE_RESULT_FALSE;
		}
		/* ラベルから実際のアドレスへ変換するアドレスを追加 */
		if(!CzMemory_add_datas(&stream->label_address, (char*)&vm_data->pc, sizeof(vm_data->pc))) {
			stream->err_code = ERR_OUTOFMEMORY;
			return PARSE_RESULT_FALSE;
		}
		/* アドレスをラベルのインデックスで仮設定 */
		if(!add_progmem(vm_data, stream, label_index)) {
			return PARSE_RESULT_FALSE;
		}
	} else {
		lex_unget_token(token);
		/* 式解析 */
		if(parse_expression(vm_data, stream) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
		if(!add_progmem(vm_data, stream, OP_JMP_NUM)) {
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* LET解析 */
static PARSE_RESULT parse_let(VM_DATA *vm_data, LEX_DATA *stream)
{
	PARSE_RESULT res;
	TOKEN *token = lex_get_token(stream);
	int var_index = 0;
	if(token->token_type == TOKEN_TYPE_ERROR) {
		stream->err_code = ERR_BUFFER_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}

	/* 変数 */
	if(strlen(token->token) != 1) {
		stream->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	if(!isalpha((int)token->token[0])) {
		stream->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	res = parse_var(token, &var_index);
	if(res == PARSE_RESULT_FALSE) {
		stream->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	/* = */
	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		stream->err_code = ERR_BUFFER_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}
	if(token->token_type != TOKEN_TYPE_EQ) {
		stream->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	/* 式解析 */
	if(parse_expression(vm_data, stream) == PARSE_RESULT_FALSE) {
		return PARSE_RESULT_FALSE;
	}

	if(!(add_progmem(vm_data, stream, OP_POP_VAR) &&
			add_progmem(vm_data, stream, var_index))) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* INPUT解析 */
static PARSE_RESULT parse_input(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN *token;
	int var_index = 0;
	PARSE_RESULT res;
	int count;

	count = 0;
	while(1) {
		/* 変数の解析 */
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		res  = parse_var(token, &var_index);
		if(res != PARSE_RESULT_TRUE) {
			stream->err_code = ERR_SYNTAX;
			return PARSE_RESULT_FALSE;
		}
		if(!(add_progmem(vm_data, stream, OP_PUSH_INT) &&
				add_progmem(vm_data, stream, var_index))) {
			return PARSE_RESULT_FALSE;
		}

		count++;

		/* カンマの解析 */
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		if(token->token_type != TOKEN_TYPE_COMMA) {
			lex_unget_token(token);
			break;
		}
	}
	if(!(add_progmem(vm_data, stream, OP_INPUT) &&
			add_progmem(vm_data, stream, count))) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* GOSUB解析 */
static PARSE_RESULT parse_gosub(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN *token;
	int label_index;

	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		stream->err_code = ERR_BUFFER_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}
	if(token->token_type == TOKEN_TYPE_MUL) { /* ラベルなら */
		if(parse_label(vm_data, stream, &label_index) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
		if(!add_progmem(vm_data, stream, OP_CALL)) {
			return PARSE_RESULT_FALSE;
		}
		/* ラベルから実際のアドレスへ変換するアドレスを追加 */
		if(!CzMemory_add_datas(&stream->label_address, (char*)&vm_data->pc, sizeof(vm_data->pc))) {
			stream->err_code = ERR_OUTOFMEMORY;
			return PARSE_RESULT_FALSE;
		}
		/* 飛び先を仮設定 */
		if(!add_progmem(vm_data, stream, label_index)) {
			return PARSE_RESULT_FALSE;
		}
	} else {
		lex_unget_token(token);
		/* 式解析 */
		if(parse_expression(vm_data, stream) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
		if(!add_progmem(vm_data, stream, OP_CALL_NUM)) {
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* RETURN解析 */
static PARSE_RESULT parse_return(VM_DATA *vm_data, LEX_DATA *stream)
{
	if(!add_progmem(vm_data, stream, OP_RET)) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* CLEAR解析 */
static PARSE_RESULT parse_clear(VM_DATA *vm_data, LEX_DATA *stream)
{
	if(!add_progmem(vm_data, stream, OP_CLEAR)) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}

/* END解析 */
static PARSE_RESULT parse_end(VM_DATA *vm_data, LEX_DATA *stream)
{
	if(!add_progmem(vm_data, stream, OP_END)) {
		return PARSE_RESULT_FALSE;
	}
	return PARSE_RESULT_TRUE;
}
/* REM解析 */
static PARSE_RESULT parse_rem(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN *token;

	/* 改行までスキップ */
	do {
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
	} while(token->token_type != TOKEN_TYPE_NL &&
			token->token_type != TOKEN_TYPE_EOF);
	lex_unget_token(token);
	return PARSE_RESULT_TRUE;
}

/* 式解析 */
static PARSE_RESULT parse_expression(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN *token;
	TOKEN_TYPE backup_type;
	enum opecode ope;

	/* 項1解析 */
	if(parse_item1(vm_data, stream) == PARSE_RESULT_FALSE) {
		return PARSE_RESULT_FALSE;
	}

	while(1) {
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		if(token->token_type != TOKEN_TYPE_EQ &&
			token->token_type != TOKEN_TYPE_LT && 
			token->token_type != TOKEN_TYPE_GT &&
			token->token_type != TOKEN_TYPE_LE &&
			token->token_type != TOKEN_TYPE_GE &&
			token->token_type != TOKEN_TYPE_NE) {
			lex_unget_token(token);
			break;
		}
		backup_type = token->token_type;
		
		/* 項1解析 */
		if(parse_item1(vm_data, stream) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}

		/* オペコードを追加 */
		ope = search_mark_to_opecode(backup_type);
		if(ope == OP_NOP) {
			stream->err_code = ERR_SYNTAX;
			return PARSE_RESULT_FALSE;
		}
		if(!add_progmem(vm_data, stream, ope)) {
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* 項1解析 */
static PARSE_RESULT parse_item1(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN_TYPE backup_type;
	enum opecode ope;

	/* 項2解析 */
	if(parse_item2(vm_data, stream) == PARSE_RESULT_FALSE) {
		return PARSE_RESULT_FALSE;
	}
	while(1) {
		TOKEN *token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		if(token->token_type != TOKEN_TYPE_PLUS &&
			token->token_type != TOKEN_TYPE_MINUS) {
			lex_unget_token(token);
			break;
		}
		backup_type = token->token_type;
	
		/* 項2解析 */
		if(parse_item2(vm_data, stream) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}

		ope = search_mark_to_opecode(backup_type);
		if(ope == OP_NOP) {
			stream->err_code = ERR_SYNTAX;
			return PARSE_RESULT_FALSE;
		}

		/* プログラムメモリへ追加 */
		if(!add_progmem(vm_data, stream, ope)) {
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* 項2解析 */
static PARSE_RESULT parse_item2(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN_TYPE backup_type;
	enum opecode ope;

	/* 基本式解析 */
	if(parse_basis_expression(vm_data, stream) == PARSE_RESULT_FALSE) {
		return PARSE_RESULT_FALSE;
	}
	while(1) {
		TOKEN *token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		if(token->token_type != TOKEN_TYPE_MUL &&
			token->token_type != TOKEN_TYPE_DIV) {
			lex_unget_token(token);
			break;
		}
		backup_type = token->token_type;
	
		/* 基本式解析 */
		if(parse_basis_expression(vm_data, stream) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}

		ope = search_mark_to_opecode(backup_type);
		if(ope == OP_NOP) {
			stream->err_code = ERR_SYNTAX;
			return PARSE_RESULT_FALSE;
		}

		/* プログラムメモリへ追加 */
		if(!add_progmem(vm_data, stream, ope)) {
			return PARSE_RESULT_FALSE;
		}
	}

	return PARSE_RESULT_TRUE;
}

/* 基本式解析 */
static PARSE_RESULT parse_basis_expression(VM_DATA *vm_data, LEX_DATA *stream)
{
	TOKEN *token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		stream->err_code = ERR_BUFFER_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}

	if(token->token_type == TOKEN_TYPE_LPARENTHESES) { /* "("なら */
		/* 式解析 */
		if(parse_expression(vm_data, stream) == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
		
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}

		if(token->token_type != TOKEN_TYPE_RPARENTHESES) { /* ")"でないなら */
			stream->err_code = ERR_SYNTAX;
			return PARSE_RESULT_FALSE;
		}
	} else {
		/* 整数へ変換 */
		PARSE_RESULT ret;
		int retval;
		lex_unget_token(token);
		ret = parse_int(stream, &retval);
		if(ret == PARSE_RESULT_FALSE) { /* 整数でないなら */
			int var_index;
			/* 変数 */
			if(parse_var(token, &var_index) == PARSE_RESULT_FALSE) {
				stream->err_code = ERR_SYNTAX;
				return PARSE_RESULT_FALSE;
			}
			if(!(add_progmem(vm_data, stream, OP_PUSH_VAR) &&
					add_progmem(vm_data, stream, var_index))) {
				return PARSE_RESULT_FALSE;
			}
		} else if(ret == PARSE_RESULT_TRUE) {	/* 整数なら */
			if(!(add_progmem(vm_data, stream, OP_PUSH_INT) &&
					add_progmem(vm_data, stream, retval))) {
				return PARSE_RESULT_FALSE;
			}
		} else {
			stream->err_code = ERR_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}

/* 整数へ変換 */
PARSE_RESULT parse_int(LEX_DATA *stream, int *retval)
{
	TOKEN *token;
	int val;
	PARSE_RESULT res;

	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		return PARSE_RESULT_FALSE;
	}
	if(token->token_type == TOKEN_TYPE_MINUS) { /* "-"なら */
		*retval = -1;
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			return PARSE_RESULT_FALSE;
		}
	} else if(token->token_type == TOKEN_TYPE_PLUS) { /* "+"なら */
		*retval = 1;
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			return PARSE_RESULT_FALSE;
		}
	} else {
		*retval = 1;
	}
	if(token->token_type == TOKEN_TYPE_EOF || token->token_type == TOKEN_TYPE_NL) {
		return PARSE_RESULT_FALSE;
	}
	res = conv_unsigned_int(token, &val);
	if(res == PARSE_RESULT_TRUE) {
		*retval = *retval * val;
	}
	
	return res;
}

/* 変数の解析　*/
static PARSE_RESULT parse_var(TOKEN *token, int *var_index)
{
	if(token->token[1] != 0) {
		return PARSE_RESULT_FALSE;
	}
	if(token->token[0] >= 'a' && token->token[1] <= 'z') {
		*var_index = token->token[0] - 'a';
		return PARSE_RESULT_TRUE;
	}
	if(token->token[0] >= 'A' && token->token[1] <= 'Z') {
		*var_index = token->token[0] - 'A';
		return PARSE_RESULT_TRUE;
	}
	return PARSE_RESULT_FALSE;
}

/* ラベルの検索 */
UINT32 search_label(VM_DATA *vm_data, char *label)
{
	size_t i;
	LABEL_TABLE *item;

	/* ラベルを検索 */
	item = (LABEL_TABLE*)vm_data->label_table.data;
	for(i = 0; i < vm_data->label_table.size / sizeof(LABEL_TABLE); i++) {
		if(Strcmp(item[i].label.data, label) == 0) {
			return i;
		}
	}

	return (UINT32)-1;
}

/* ラベルに割り当てられているアドレスを変更 */
static void change_label_address(VM_DATA *vm_data, UINT32 label_index, UINT32 address)
{
	LABEL_TABLE *item;

	item = (LABEL_TABLE*)vm_data->label_table.data;
	item[label_index].address = address;
}

/* ラベルに割り当てられているアドレスを取得 */
UINT32 get_label_address(VM_DATA *vm_data, UINT32 label_index)
{
	LABEL_TABLE *item;

	item = (LABEL_TABLE*)vm_data->label_table.data;
	return item[label_index].address;
}

/* ラベルの追加 */
static UINT32 add_label(VM_DATA *vm_data, char *label)
{
	size_t len;
	LABEL_TABLE label_table;
	UINT32 label_index;

	len = strlen(label) + 1;

	/* すでに登録済かどうか検索 */
	label_index = search_label(vm_data, label);
	if(label_index != (UINT32)-1) { /* 登録済なら */
		return label_index;
	}

	/* 追加登録 */
	label_index = vm_data->label_table.size / sizeof(LABEL_TABLE);
	Label_table_init(&label_table);
	if(!CzString_add_datas(&label_table.label, label, len)) {
		Label_table_delete(&label_table);
		return (UINT32)-1;
	}
	if(!CzMemory_add_datas(&vm_data->label_table, (char*)&label_table, sizeof(label_table))) {
		Label_table_delete(&label_table);
		return (UINT32)-1;
	}
	return label_index;
}

/* ラベル解析 */
static PARSE_RESULT parse_label(VM_DATA *vm_data, LEX_DATA *stream, int *index)
{
	TOKEN *token;
	UINT32 label_index;

	token = lex_get_token(stream);
	if(token->token_type == TOKEN_TYPE_ERROR) {
		stream->err_code = ERR_BUFFER_OVERFLOW;
		return PARSE_RESULT_FALSE;
	}
	if(!isalpha(token->token[0])) {	/* アルファベットで始まっていないなら */
		stream->err_code = ERR_SYNTAX;
		return PARSE_RESULT_FALSE;
	}
	/* ラベルを追加 */
	label_index = add_label(vm_data, token->token);
	if(label_index == (UINT32)-1) {
		stream->err_code = ERR_OUTOFMEMORY;
		return PARSE_RESULT_FALSE;
	}
	*index = label_index;
	return PARSE_RESULT_TRUE;
}

/* 命令リスト解析 */
static PARSE_RESULT parse_statement_list(VM_DATA *vm_data, LEX_DATA *stream)
{
	PARSE_RESULT res;
	TOKEN *token;

	/* 命令解析 */
	res = parse_statement(vm_data, stream);
	if(res == PARSE_RESULT_FALSE) {
		return PARSE_RESULT_FALSE;
	}
	
	while(1) {
		token = lex_get_token(stream);
		if(token->token_type == TOKEN_TYPE_ERROR) {
			stream->err_code = ERR_BUFFER_OVERFLOW;
			return PARSE_RESULT_FALSE;
		}
		/* コロン(:)チェック */
		if(token->token_type != TOKEN_TYPE_COLON) {
			lex_unget_token(token);
			break;
		}
		/* 命令解析 */
		res = parse_statement(vm_data, stream);
		if(res == PARSE_RESULT_FALSE) {
			return PARSE_RESULT_FALSE;
		}
	}
	return PARSE_RESULT_TRUE;
}
